#pragma once

#include <filesystem>
#include <string>
#include <sstream>
#include <vector>

namespace dumb {

	struct Token {
		enum class Type {
			KEYWORD,
			IDENTIFIER,
			OPERATOR,
			INTEGER_LITERAL,
			FLOAT_LITERAL,
			STRING_LITERAL,
			SEMICOLON,
			COMMA,
			OPEN_PAREN,
			CLOSE_PAREN,
			OPEN_BRACE,
			CLOSE_BRACE,
			OPEN_BRACKET,
			CLOSE_BRACKET
		};

		Type type;
		std::string data;
		size_t line;
		size_t column;

		Token(Type type, std::string data, size_t line, size_t column) : type(type), data(std::move(data)), line(line), column(column) {}
	};

	class Lexer {
		private:
			std::string sourceCode;
			std::vector<Token> tokenStream;
			std::stringstream buffer;
			size_t line;
			size_t column;
			size_t pointer;

			static const std::vector<std::string> Keywords;

			void resetBuffer() {
				this->buffer.str("");
				this->buffer.clear();
			}

			const char& current() const {
				return this->peek(0);
			}

			const char& next() const {
				return this->peek(1);
			}

			const char& peek(size_t amount = 0) const {
				size_t address = this->pointer + amount;
				return address < this->sourceCode.size() ? this->sourceCode.at(address) : Lexer::EoF;
			}

			const char& consume() {
				const char& current = this->peek();
				this->pointer++;
				this->column++;

				return current;
			}

			static bool IsOperator(const char& car) {
				switch (car) {
					case '+':
					case '-':
					case '*':
					case '/':
					case '=':
					case ':':
					case '.':
						return true;
					default:
						return false;
				}
			}

			static bool IsKeyword(const std::string& identifier) {
				return std::any_of(Lexer::Keywords.begin(), Lexer::Keywords.end(), [&identifier](std::string keyword){
					return keyword == identifier;
				});
			}

			void readIdentifier();
			void readOperator();
			void readStringLiteral();
			void readNumber();
		public:
			constexpr static char EoF = '\0';
			constexpr static char EOL = '\n';
			explicit Lexer(std::filesystem::path sourceFile);

			void tokenize();
	};

}

